"""Module to represent a package."""
from constants import PackageManagers

class MetaPackage:
    """Class to represent a package."""
    instances = []

    def __init__(self, pkgname, pkgtype=None, pkgorg=None):
        self.instances.append(self) # adding the instance to colllective
        if len(pkgname.split(':')) == 2:
            if pkgtype == PackageManagers.MAVEN.value:
                if pkgorg is None:
                    self._pkg_name = pkgname.split(':')[1]
                    self._org_id = pkgname.split(':')[0]
        else:
            self._pkg_name = pkgname
            self._org_id = pkgorg
        self._exists = None
        self._pkg_type = pkgtype
        self._score = None
        self._timestamp = None
        self._version_count = None
        self._fork_count = None
        self._subs_count = None
        self._star_count = None
        self._contributor_count = None
        self._download_count = None
        self._issue_count = None
        #self._pkg_ver = pkgver TBA
        self._risk_missing = None
        self._risk_low_score = None
        self._risk_min_versions = None
        self._risk_too_new = None

    def __repr__(self):
        return self._pkg_name

    def __str__(self):
        return str(self._pkg_name)

    def listall(self):
        """List all the attributes of the class.

        Returns:
            list: List of all the attributes of the class.
        """
        lister = []
        lister.append(self._pkg_name)
        lister.append(self._pkg_type)
        lister.append(self._exists)
        lister.append(self._org_id)
        lister.append(self._score)
        lister.append(self._version_count)
        lister.append(self._timestamp)
        lister.append(self._risk_missing)
        lister.append(self._risk_low_score)
        lister.append(self._risk_min_versions)
        lister.append(self._risk_too_new)
        lister.append(self.has_risk())
        return lister

    @staticmethod
    def get_instances():
        """Get all instances of the class.

        Returns:
            list: List of all instances of the class.
        """
        return MetaPackage.instances

    @property
    def pkg_name(self):
        """Property for the package name.

        Returns:
            str: Package name.
        """
        return self._pkg_name


    @property
    def pkg_type(self):
        """Property for the package type.

        Returns:
            str: Package type.
        """
        return self._pkg_type

    @pkg_type.setter
    def pkg_type(self, pkg_type):
        self._pkg_type = pkg_type

    @property
    def author(self):
        """Property for the author.

        Returns:
            str: Author.
        """
        return self._author

    @author.setter
    def author(self, a):
        self._author = a

    @property
    def author_email(self):
        """Property for the author email.

        Returns:
            str: Author email.
        """
        return self._author_email

    @author_email.setter
    def author_email(self, a):
        self._author_email = a

    @property
    def exists(self):
        """Property defining if the package exists.

        Returns:
            boolean: True if the package exists, False otherwise.
        """
        return self._exists

    @exists.setter
    def exists(self, a):
        self._exists = a

    @property
    def publisher(self):
        """Property for the publisher.

        Returns:
            str: Publisher.
        """
        return self._publisher

    @publisher.setter
    def publisher(self, a):
        self._publisher = a

    @property
    def publisher_email(self):
        """Property for the publisher email.

        Returns:
            str: Publisher email.
        """
        return self._publisher_email

    @publisher.setter
    def publisher(self, a):
        self._publisher_email = a

    @property
    def maintainer(self):
        """Property for the maintainer.

        Returns:
            str: Maintainer.
        """
        return self._maintainer

    @maintainer.setter
    def maintainer(self, a):
        self._maintainer = a

    @property
    def maintainer_email(self):
        """Property for the maintainer email.

        Returns:
            str: Maintainer email.
        """
        return self._maintainer_email

    @maintainer_email.setter
    def maintainer_email(self, email_address):
        self._maintainer_email = email_address

    @property
    def fork_count(self):
        """Property for the fork count.

        Returns:
            int: Fork count.
        """
        return self._fork_count

    @fork_count.setter
    def fork_count(self, count):
        self._fork_count = count

    @property
    def subs_count(self):
        """Property for the subscription count.

        Returns:
            int: Subscription count.
        """
        return self._subs_count

    @subs_count.setter
    def subs_count(self, a):
        self._subs_count = a

    @property
    def star_count(self):
        """Property for the star count.

        Returns:
            int: Star count.
        """
        return self._star_count

    @star_count.setter
    def star_count(self, a):
        self._star_count = a

    @property
    def download_count(self):
        """Property for the download count.

        Returns:
            int: Download count.
        """
        return self._download_count

    @download_count.setter
    def download_count(self, count):
        self._download_count = count

    @property
    def score(self):
        """Property for the score.

        Returns:
            int: Score.
        """
        return self._score

    @score.setter
    def score(self, a):
        self._score = a

    @property
    def dependencies(self):
        """Property for the dependencies.

        Returns:
            list: List of dependencies.
        """
        return self._dependencies

    @dependencies.setter
    def dependencies(self, dependency_list):
        self._dependencies = dependency_list

    @property
    def issue_count(self):
        """Property for the issue count.

        Returns:
            int: Issue count.
        """
        return self._issue_count

    @issue_count.setter
    def issue_count(self, count):
        self._issue_count = count

    @property
    def risk_missing(self):
        """Risk property for missing package.

        Returns:
            bool: True if the package is missing, False otherwise.
        """
        return self._risk_missing

    @risk_missing.setter
    def risk_missing(self, is_missing):
        self._risk_missing = is_missing

    @property
    def risk_low_score(self):
        """Risk property for having a low score

        Returns:
            bool: True if the package has a low score, False otherwise.
        """
        return self._risk_low_score

    @risk_low_score.setter
    def risk_low_score(self, is_low_score):
        self._risk_low_score = is_low_score

    @property
    def risk_min_versions(self):
        """Risk property for too few versions

        Returns:
            bool: True if the package has too few versions, False otherwise.
        """
        return self._risk_min_versions

    @risk_min_versions.setter
    def risk_min_versions(self, is_risk_min_versions):
        self._risk_min_versions = is_risk_min_versions

    @property
    def risk_too_new(self):
        """Risk property for too new package

        Returns:
            bool: True if the package is too new, False otherwise.
        """
        return self._risk_too_new

    @risk_too_new.setter
    def risk_too_new(self, is_risk_too_new):
        self._risk_too_new = is_risk_too_new
        
    @property
    def contributor_count(self):
        """Property for the contributor count.

        Returns:
            int: Contributor count.
        """
        return self._contributor_count

    @contributor_count.setter
    def contributor_count(self, a):
        self._contributor_count = a

    @property
    def org_id(self):
        """Property for the organization ID.

        Returns:
            str: Organization ID.
        """
        return self._org_id

    @org_id.setter
    def org_id(self, a):
        self._org_id = a

    @property
    def version_count(self):
        """Property for the version count.

        Returns:
            int: Version count.
        """
        return self._version_count

    @version_count.setter
    def version_count(self, a):
        self._version_count = a

    @property
    def timestamp(self):
        """Property for the timestamp.

        Returns:
            timestamp: Timestamp.
        """

        return self._timestamp

    @timestamp.setter
    def timestamp(self, timestamp): #unix timestamp
        self._timestamp = timestamp

    def has_risk(self):
        """Check if the package has any risk.

        Returns:
            bool: True if the package has any risk, False otherwise.
        """
        if self._risk_missing or self._risk_low_score or self._risk_min_versions or self._risk_too_new:
            return True
        return False
# not-supported for now: hasTests, testsSize, privateRepo



"""Heuristics for package analysis."""
import time
import logging  # Added import
from constants import Constants, DefaultHeuristics

STG = f"{Constants.ANALYSIS} "

def combobulate_min(pkgs):
    """Run to check the existence of the packages in the registry.

    Args:
        pkgs (list): List of packages to check.
    """
    for x in pkgs:
        test_exists(x)

def combobulate_heur(pkgs):
    """Run heuristics on the packages.

    Args:
        pkgs (list): List of packages to check.
    """
    for x in pkgs:
        test_exists(x)
        if x.exists is True:
            test_score(x)
            test_timestamp(x)
            test_version_count(x)
    stats_exists(pkgs)

def test_exists(x):
    """Check if the package exists on the public provider.

    Args:
        x (str): Package to check.
    """
    if x.exists is True:
        logging.info("%sPackage: %s is present on public provider.", STG, x)
        x.risk_missing = False
    elif x.exists is False:
        logging.warning("%sPackage: %s is NOT present on public provider.", STG, x)
        x.risk_missing = True
    else:
        logging.info("%sPackage: %s test skipped.", STG, x)

def test_score(x):
    """Check the score of the package.

    Args:
        x (str): Package to check.
    """
    ttxt = ". Mid set to " + str(DefaultHeuristics.SCORE_THRESHOLD.value) + ")"
    if x.score is not None:
        if x.score > DefaultHeuristics.SCORE_THRESHOLD.value:
            logging.info("%s.... package scored ABOVE MID - %s%s",
                STG, str(x.score), ttxt)
            x.risk_low_score = False
        elif x.score <= DefaultHeuristics.SCORE_THRESHOLD.value and x.score > DefaultHeuristics.RISKY_THRESHOLD.value:
            logging.warning("%s.... [RISK] package scored BELOW MID - %s%s",
                STG, str(x.score), ttxt)
            x.risk_low_score = False
        elif x.score <= DefaultHeuristics.RISKY_THRESHOLD.value:
            logging.warning("%s.... [RISK] package scored LOW - %s%s", STG, str(x.score), ttxt)
            x.risk_low_score = True

def test_timestamp(x):
    """Check the timestamp of the package.

    Args:
        x (str): Package to check.
    """
    if x.timestamp is not None:
        dayspast = (time.time()*1000 - x.timestamp)/86400000
        logging.info("%s.... package is %d days old.", STG, int(dayspast))
        if dayspast < 2:  # freshness test
            logging.warning("%s.... [RISK] package is SUSPICIOUSLY NEW.", STG)
            x.risk_too_new = True
        else:
            logging.debug("%s.... package is not suspiciously new.", STG)
            x.risk_too_new = False

def stats_exists(pkgs):
    """Summarize the existence of the packages on the public provider.

    Args:
        pkgs (list): List of packages to check.
    """
    count = sum(1 for x in pkgs if x.exists is True)
    total = len(pkgs)
    percentage = (count / total) * 100 if total > 0 else 0
    logging.info("%s%d out of %d packages were present on the public provider (%.2f%% of total).",
                 STG, count, total, percentage)

def test_version_count(pkg):
    """Check the version count of the package.

    Args:
        pkg (str): Package to check.
    """
    if pkg.version_count is not None:
        if pkg.version_count < 2:
            logging.warning("%s.... [RISK] package history is SHORT. Total %d versions committed.",
                            STG, pkg.version_count)
            pkg.risk_min_versions = True
        else:
            logging.info("%s.... Total %d versions committed.", STG, pkg.version_count)
            pkg.risk_min_versions = False
    else:
        logging.warning("%s.... Package version count not available.", STG)



"""DepGate - Dependency supply-chain/confusion risk checker (hard fork)

    Raises:
        TypeError: If the input list cannot be processed

    Returns:
        int: Exit code
"""
import csv
import sys
import logging
import json

# internal module imports (kept light to avoid heavy deps on --help)
from metapackage import MetaPackage as metapkg
from constants import ExitCodes, PackageManagers, Constants
from args import parse_args

SUPPORTED_PACKAGES = Constants.SUPPORTED_PACKAGES

def load_pkgs_file(file_name):
    """Loads the packages from a file.

    Args:
        file_name (str): File path containing the list of packages.

    Raises:
        TypeError: If the input list cannot be processed

    Returns:
        list: List of packages
    """
    try:
        with open(file_name, encoding='utf-8') as file:
            return [line.strip() for line in file]
    except FileNotFoundError as e:
        logging.error("File not found: %s, aborting", e)
        sys.exit(ExitCodes.FILE_ERROR.value)
    except IOError as e:
        logging.error("IO error: %s, aborting", e)
        sys.exit(ExitCodes.FILE_ERROR.value)

def scan_source(pkgtype, dir_name, recursive=False):
    """Scans the source directory for packages.

    Args:
        pkgtype (str): Package manager type, i.e. "npm".
        dir (str): Directory path to scan.
        recursive (bool, optional): Option to recurse into subdirectories. Defaults to False.

    Returns:
        list: List of packages found in the source directory.
    """
    if pkgtype == PackageManagers.NPM.value:
        from registry import npm as _npm
        return _npm.scan_source(dir_name, recursive)
    elif pkgtype == PackageManagers.MAVEN.value:
        from registry import maven as _maven
        return _maven.scan_source(dir_name, recursive)
    elif pkgtype == PackageManagers.PYPI.value:
        from registry import pypi as _pypi
        return _pypi.scan_source(dir_name, recursive)
    else:
        logging.error("Selected package type doesn't support import scan.")
        sys.exit(ExitCodes.FILE_ERROR.value)

def check_against(check_type, level, check_list):
    """Checks the packages against the registry.

    Args:
        check_type (str): Package manager type, i.e. "npm".
        check_list (list): List of packages to check.
    """


    if check_type == PackageManagers.NPM.value:
        # Only fetch details for levels 1 and 2
        should_fetch_details = level in (Constants.LEVELS[2], Constants.LEVELS[3])
        from registry import npm as _npm
        _npm.recv_pkg_info(check_list, should_fetch_details)
    elif check_type == PackageManagers.MAVEN.value:
        from registry import maven as _maven
        _maven.recv_pkg_info(check_list)
    elif check_type == PackageManagers.PYPI.value:
        from registry import pypi as _pypi
        _pypi.recv_pkg_info(check_list)
    else:
        logging.error("Selected package type doesn't support registry check.")
        sys.exit(ExitCodes.FILE_ERROR.value)

def export_csv(instances, path):
    """Exports the package properties to a CSV file.

    Args:
        instances (list): List of package instances.
        path (str): File path to export the CSV.
    """
    headers = ["Package Name","Package Type", "Exists on External",
            "Org/Group ID","Score","Version Count","Timestamp",
            "Risk: Missing", "Risk: Low Score","Risk: Min Versions","Risk: Too New", "Risk: Any Risks"]
    rows = [headers]
    for x in instances:
        rows.append(x.listall())
    try:
        with open(path, 'w', newline='', encoding='utf-8') as file:
            export = csv.writer(file)
            export.writerows(rows)
        logging.info("CSV file has been successfully exported at: %s", path)
    except (OSError, csv.Error) as e:
        logging.error("CSV file couldn't be written to disk: %s", e)
        sys.exit(1)

def export_json(instances, path):
    """Exports the package properties to a JSON file.

    Args:
        instances (list): List of package instances.
        path (str): File path to export the JSON.
    """
    data = []
    for x in instances:
        data.append({
            "packageName": x.pkg_name,
            "orgId": x.org_id,
            "packageType": x.pkg_type,
            "exists": x.exists,
            "score": x.score,
            "versionCount": x.version_count,
            "createdTimestamp": x.timestamp,
            "risk": {
                "hasRisk": x.has_risk(),
                "isMissing": x.risk_missing,
                "hasLowScore": x.risk_low_score,
                "minVersions": x.risk_min_versions,
                "isNew": x.risk_too_new
            }
        })
    try:
        with open(path, 'w', encoding='utf-8') as file:
            json.dump(data, file, ensure_ascii=False, indent=4)
        logging.info("JSON file has been successfully exported at: %s", path)
    except OSError as e:
        logging.error("JSON file couldn't be written to disk: %s", e)
        sys.exit(1)

def main():
    """Main function of the program."""
    # the most important part of any program starts here

    args = parse_args()

    # Configure logging
    log_level = getattr(logging, args.LOG_LEVEL.upper(), logging.INFO)


    if '-h' in sys.argv or '--help' in sys.argv:
        # Ensure help output is always at INFO level
        logging.basicConfig(level=logging.INFO, format=Constants.LOG_FORMAT)
    else:
        if args.LOG_FILE:
            logging.basicConfig(filename=args.LOG_FILE, level=log_level,
                                format=Constants.LOG_FORMAT)  # Used LOG_FORMAT constant
        else:
            # If log is not set to a LOG_FILE and quiet mode is not enabled, set log level to none
            if args.QUIET:
                logging.disable(logging.CRITICAL)
            else:
                logging.basicConfig(level=log_level,
                                format=Constants.LOG_FORMAT)  # Used LOG_FORMAT constant

    logging.info("Arguments parsed.")

    logging.info(r"""
┬─┐ ┬─┐ ┬─┐ ┌─┐ ┬─┐ ┌┐┐ ┬─┐
│ │ │─  │─┘ │ ┬ │─┤  │  │─
──┘ ┴─┘ ┴   │─┘ ┘ │  ┘  ┴─┘

  Dependency Supply-Chain/Confusion Risk Checker
""")

    # SCAN & FLAG ARGS

    # Check if recursive option is used without directory
    if args.RECURSIVE and not args.FROM_SRC:
        logging.warning("Recursive option is only applicable to source scans.")

    #IMPORT
    pkglist = []
    if args.LIST_FROM_FILE:
        pkglist = load_pkgs_file(args.LIST_FROM_FILE[0])
    elif args.FROM_SRC:
        pkglist = scan_source(args.package_type, args.FROM_SRC[0], recursive=args.RECURSIVE)
    elif args.SINGLE:
        pkglist.append(args.SINGLE[0])

    if not pkglist or not isinstance(pkglist, list):
        logging.warning("No packages found in the input list.")
        sys.exit(ExitCodes.SUCCESS.value)

    logging.info("Package list imported: %s", str(pkglist))

    if args.package_type == PackageManagers.NPM.value:
        for pkg in pkglist:
            metapkg(pkg, args.package_type)
    elif args.package_type == PackageManagers.MAVEN.value:
        for pkg in pkglist: # format org_id:package_id
            metapkg(pkg.split(':')[1], args.package_type, pkg.split(':')[0])
    elif args.package_type == PackageManagers.PYPI.value:
        for pkg in pkglist:
            metapkg(pkg, args.package_type)

    # QUERY & POPULATE
    check_against(args.package_type, args.LEVEL, metapkg.instances)

    # ANALYZE
    if args.LEVEL in (Constants.LEVELS[0], Constants.LEVELS[1]):
        from analysis import heuristics as _heur
        _heur.combobulate_min(metapkg.instances)
    elif args.LEVEL in (Constants.LEVELS[2], Constants.LEVELS[3]):
        from analysis import heuristics as _heur
        _heur.combobulate_heur(metapkg.instances)

    # OUTPUT
    if args.CSV:
        export_csv(metapkg.instances, args.CSV)
    if args.JSON:
        export_json(metapkg.instances, args.JSON)

    # Check if any package was not found
    has_risk = any( x.has_risk() for x in metapkg.instances)
    if has_risk:
        logging.warning("One or more packages have identified risks.")
        if args.ERROR_ON_WARNINGS:
            logging.error("Warnings present, exiting with non-zero status code.")
            sys.exit(ExitCodes.EXIT_WARNINGS.value)

    sys.exit(ExitCodes.SUCCESS.value)

if __name__ == "__main__":
    main()



"""
  NPM registry module. This module is responsible for checking
  the existence of the packages in the NPM registry and scanning'
  the source code for dependencies."""
import json
import sys
import os
import time
from datetime import datetime as dt
import logging  # Added import
import requests
from constants import ExitCodes, Constants

def get_keys(data):
    """Get all keys from a nested dictionary.

    Args:
        data (dict): Dictionary to extract keys from.

    Returns:
        list: List of all keys in the dictionary.
    """
    result = []
    for key in data.keys():
        if not isinstance(data[key], dict):
            result.append(key)
        else:
            result += get_keys(data[key])
    return result

def get_package_details(pkg, url):
    """Get the details of a package from the NPM registry.

    Args:
        x (_type_): _description_
        url (_type_): _description_
    """

    # Short sleep to avoid rate limiting
    time.sleep(0.1)

    try:
        logging.debug("Checking package: %s", pkg.pkg_name)
        package_url = url + pkg.pkg_name
        package_headers = {
            'Accept': 'application/vnd.npm.install-v1+json; q=1.0, application/json; q=0.8, */*'}
        res = requests.get(package_url,
            headers=package_headers,
            timeout=Constants.REQUEST_TIMEOUT)
    except requests.Timeout:
        logging.error("Request timed out after %s seconds", Constants.REQUEST_TIMEOUT)
        sys.exit(ExitCodes.CONNECTION_ERROR.value)
    except requests.RequestException as e:
        logging.error("Connection error: %s", e)
        sys.exit(ExitCodes.CONNECTION_ERROR.value)
    if res.status_code == 404:
        pkg.exists = False
        return
    try:
        package_info = json.loads(res.text)
    except json.JSONDecodeError:
        logging.warning("Couldn't decode JSON, assuming package missing.")
        pkg.exists = False
        return
    pkg.exists = True
    pkg.version_count = len(package_info['versions'])

def recv_pkg_info(pkgs, should_fetch_details=False, details_url=Constants.REGISTRY_URL_NPM, url=Constants.REGISTRY_URL_NPM_STATS):
    """Check the existence of the packages in the NPM registry.

    Args:
        pkgs (list): List of packages to check.
        url (str, optional): NPM Url. Defaults to Constants.REGISTRY_URL_NPM.
    """
    logging.info("npm checker engaged.")
    pkg_list = []
    for pkg in pkgs:
        pkg_list.append(pkg.pkg_name)
        if should_fetch_details:
            get_package_details(pkg, details_url)
    payload =  '['+','.join(f'"{w}"' for w in pkg_list)+']' #list->payload conv
    headers = { 'Accept': 'application/json',
                'Content-Type': 'application/json'}
    logging.info("Connecting to registry at %s ...", url)
    try:
        res = requests.post(url, data=payload, headers=headers,
                          timeout=Constants.REQUEST_TIMEOUT)
        if res.status_code != 200:
            logging.error("Unexpected status code (%s)", res.status_code)
            sys.exit(ExitCodes.CONNECTION_ERROR.value)
        pkg = json.loads(res.text)
    except requests.Timeout:
        logging.error("Request timed out after %s seconds", Constants.REQUEST_TIMEOUT)
        sys.exit(ExitCodes.CONNECTION_ERROR.value)
    except requests.RequestException as e:
        logging.error("Connection error: %s", e)
        sys.exit(ExitCodes.CONNECTION_ERROR.value)
    for i in pkgs:
        if i.pkg_name in pkg:
            package_info = pkg[i.pkg_name]
            i.exists = True
            i.score = package_info.get('score', {}).get('final', 0)
            timex = package_info.get('collected', {}).get('metadata', {}).get('date', '')
            fmtx ='%Y-%m-%dT%H:%M:%S.%fZ'
            try:
                unixtime = int(dt.timestamp(dt.strptime(timex, fmtx))*1000)
                i.timestamp = unixtime
            except ValueError as e:
                logging.warning("Couldn't parse timestamp: %s", e)
                i.timestamp = 0
        else:
            i.exists = False


def scan_source(dir_name, recursive=False):
    """Scan the source code for dependencies.

    Args:
        dir_name (str): Directory to scan.
        recursive (bool, optional): _description_. Defaults to False.

    Returns:
        list: List of dependencies found in the source code.
    """
    try:
        logging.info("npm scanner engaged.")
        pkg_files = []
        if recursive:
            for root, _, files in os.walk(dir_name):
                if Constants.PACKAGE_JSON_FILE in files:
                    pkg_files.append(os.path.join(root, Constants.PACKAGE_JSON_FILE))
        else:
            path = os.path.join(dir_name, Constants.PACKAGE_JSON_FILE)
            if os.path.isfile(path):
                pkg_files.append(path)
            else:
                logging.error("package.json not found, unable to continue.")
                sys.exit(ExitCodes.FILE_ERROR.value)

        lister = []
        for path in pkg_files:
            with open(path, "r", encoding="utf-8") as file:
                body = file.read()
            filex = json.loads(body)
            lister.extend(list(filex.get('dependencies', {}).keys()))
            if 'devDependencies' in filex:
                lister.extend(list(filex['devDependencies'].keys()))
        return list(set(lister))
    except (FileNotFoundError, IOError, json.JSONDecodeError) as e:
        logging.error("Couldn't import from given path, error: %s", e)
        sys.exit(ExitCodes.FILE_ERROR.value)



"""Maven registry interaction module."""
import json
import os
import sys
import time
import logging
import xml.etree.ElementTree as ET
import requests
from constants import ExitCodes, Constants

def recv_pkg_info(pkgs, url=Constants.REGISTRY_URL_MAVEN):
    """Check the existence of the packages in the Maven registry.

    Args:
        pkgs (list): List of packages to check.
        url (str, optional): Maven Url. Defaults to Constants.REGISTRY_URL_MAVEN.
    """
    logging.info("Maven checker engaged.")
    payload = {"wt": "json", "rows": 20}
    #TODO move everything off names and modify instances instead
    for x in pkgs:
        tempstring = "g:" + x.org_id + " a:" + x.pkg_name
        payload.update({"q": tempstring})
        #print(payload) 
        headers = { 'Accept': 'application/json',
                'Content-Type': 'application/json'}
        try:
            # Sleep to avoid rate limiting
            time.sleep(0.1)
            res = requests.get(url, params=payload, headers=headers, 
                             timeout=Constants.REQUEST_TIMEOUT)
        except requests.Timeout:
            logging.error("Request timed out after %s seconds", Constants.REQUEST_TIMEOUT)
            sys.exit(ExitCodes.CONNECTION_ERROR.value)
        except requests.RequestException as e:
            logging.error("Connection error: %s", e)
            sys.exit(ExitCodes.CONNECTION_ERROR.value)

        j = json.loads(res.text)
        number_found = j.get('response', {}).get('numFound', 0)
        if number_found == 1: #safety, can't have multiples
            x.exists = True
            x.timestamp = j.get('response', {}).get('docs', [{}])[0].get('timestamp', 0)
            x.version_count = j.get('response', {}).get('docs', [{}])[0].get('versionCount', 0)
        elif number_found > 1:
            logging.warning("Multiple packages found, skipping")
            x.exists = False
        else:
            x.exists = False

def scan_source(dir_name, recursive=False):
    """Scan the source directory for pom.xml files.

    Args:
        dir_name (str): Directory to scan.
        recursive (bool, optional): Whether to scan recursively. Defaults to False.

    Returns:
        _type_: _description_
    """
    try:
        logging.info("Maven scanner engaged.")
        pom_files = []
        if recursive:
            for root, _, files in os.walk(dir_name):
                if Constants.POM_XML_FILE in files:
                    pom_files.append(os.path.join(root, Constants.POM_XML_FILE))
        else:
            path = os.path.join(dir_name, Constants.POM_XML_FILE)
            if os.path.isfile(path):
                pom_files.append(path)
            else:
                logging.error("pom.xml not found. Unable to scan.")
                sys.exit(ExitCodes.FILE_ERROR.value)

        lister = []
        for path in pom_files:
            tree = ET.parse(path)
            pom = tree.getroot()
            ns = ".//{http://maven.apache.org/POM/4.0.0}"
            for dependencies in pom.findall(ns + 'dependencies'):
                for dependency in dependencies.findall(ns + 'dependency'):
                    group = dependency.find(ns + 'groupId').text
                    artifact = dependency.find(ns + 'artifactId').text
                    lister.append(group + ':' + artifact)
        return list(set(lister))
    except (FileNotFoundError, ET.ParseError) as e:
        logging.error("Couldn't import from given path, error: %s", e)
        sys.exit(ExitCodes.FILE_ERROR.value)



"""Constants used in the project."""

from enum import Enum


class ExitCodes(Enum):
    """Exit codes for the program.

    Args:
        Enum (int): Exit codes for the program.
    """

    SUCCESS = 0
    CONNECTION_ERROR = 2
    FILE_ERROR = 1
    EXIT_WARNINGS = 3


class PackageManagers(Enum):
    """Package managers supported by the program.

    Args:
        Enum (string): Package managers supported by the program.
    """

    NPM = "npm"
    PYPI = "pypi"
    MAVEN = "maven"

class DefaultHeuristics(Enum):
    """Default heuristics for the program.

    Args:
        Enum (int): Default heuristics for the program.
    """

    MIN_VERSIONS = 2
    NEW_DAYS_THRESHOLD = 2
    SCORE_THRESHOLD = 0.6
    RISKY_THRESHOLD = 0.15

class Constants:
    """General constants used in the project."""

    REGISTRY_URL_PYPI = "https://pypi.org/pypi/"
    REGISTRY_URL_NPM = "https://registry.npmjs.org/"
    REGISTRY_URL_NPM_STATS = "https://api.npms.io/v2/package/mget"
    REGISTRY_URL_MAVEN = "https://search.maven.org/solrsearch/select"
    SUPPORTED_PACKAGES = [
        PackageManagers.NPM.value,
        PackageManagers.PYPI.value,
        PackageManagers.MAVEN.value,
    ]
    LEVELS = ["compare", "comp", "heuristics", "heur"]
    REQUIREMENTS_FILE = "requirements.txt"
    PACKAGE_JSON_FILE = "package.json"
    POM_XML_FILE = "pom.xml"
    LOG_FORMAT = "[%(levelname)s] %(message)s"  # Added LOG_FORMAT constant
    ANALYSIS = "[ANALYSIS]"
    REQUEST_TIMEOUT = 30  # Timeout in seconds for all HTTP requests


"""PyPI registry module."""
import json
import sys
import os
import time
from datetime import datetime as dt
import logging  # Added import
import requests
import requirements
from constants import ExitCodes, Constants

def recv_pkg_info(pkgs, url=Constants.REGISTRY_URL_PYPI):
    """Check the existence of the packages in the PyPI registry.

    Args:
        pkgs (list): List of packages to check.
        url (str, optional): Url for PyPi. Defaults to Constants.REGISTRY_URL_PYPI.
    """
    logging.info("PyPI registry engaged.")
    payload = {}
    for x in pkgs:
        # Sleep to avoid rate limiting
        time.sleep(0.1)
        fullurl = url + x.pkg_name + '/json'
        logging.debug(fullurl)
        headers = {'Accept': 'application/json',
                   'Content-Type': 'application/json'}
        try:
            res = requests.get(fullurl, params=payload, headers=headers, 
                             timeout=Constants.REQUEST_TIMEOUT)
        except requests.Timeout:
            logging.error("Request timed out after %s seconds", Constants.REQUEST_TIMEOUT)
            exit(ExitCodes.CONNECTION_ERROR.value)
        except requests.RequestException as e:
            logging.error("Connection error: %s", e)
            exit(ExitCodes.CONNECTION_ERROR.value)
        if res.status_code == 404:
            # Package not found
            x.exists = False
            continue
        if res.status_code != 200:
            logging.error("Connection error, status code: %s", res.status_code)
            exit(ExitCodes.CONNECTION_ERROR.value)
        try:
            j = json.loads(res.text)
        except json.JSONDecodeError:
            logging.warning("Couldn't decode JSON, assuming package missing.")
            x.exists = False
            continue
        if j['info']:
            x.exists = True
            latest = j['info']['version']
            for version in j['releases']:
                if version == latest:
                    timex = j['releases'][version][0]['upload_time_iso_8601']
                    fmtx = '%Y-%m-%dT%H:%M:%S.%fZ'
                    try:
                        unixtime = int(dt.timestamp(dt.strptime(timex, fmtx)) * 1000)
                        x.timestamp = unixtime
                    except ValueError as e:
                        logging.warning("Couldn't parse timestamp %s, setting to 0.", e)
                        x.timestamp = 0
            x.version_count = len(j['releases'])
        else:
            x.exists = False

def scan_source(dir_name, recursive=False):
    """Scan the source directory for requirements.txt files.

    Args:
        dir_name (str): Directory to scan.
        recursive (bool, optional): Whether to recurse into subdirectories. Defaults to False.

    Raises:
        FileNotFoundError: _description_

    Returns:
        _type_: _description_
    """
    try:
        logging.info("PyPI scanner engaged.")
        req_files = []
        if recursive:
            for root, _, files in os.walk(dir_name):
                if Constants.REQUIREMENTS_FILE in files:
                    req_files.append(os.path.join(root, Constants.REQUIREMENTS_FILE))
        else:
            path = os.path.join(dir_name, Constants.REQUIREMENTS_FILE)
            if os.path.isfile(path):
                req_files.append(path)
            else:
                logging.error("requirements.txt not found, unable to continue.")
                sys.exit(ExitCodes.FILE_ERROR.value)

        all_requirements = []
        for path in req_files:
            with open(path, "r", encoding="utf-8") as file:
                body = file.read()
            reqs = requirements.parse(body)
            all_requirements.extend([x.name for x in reqs])
        return list(set(all_requirements))
    except (FileNotFoundError, IOError) as e:
        logging.error("Couldn't import from given path '%s', error: %s", path, e)
        sys.exit(ExitCodes.FILE_ERROR.value)



"""Argument parsing functionality for DepGate (hard fork)."""

import argparse
from constants import Constants

def parse_args():
    """Parses the arguments passed to the program."""
    parser = argparse.ArgumentParser(
        prog="depgate.py",
        description="DepGate - Dependency supply-chain risk and confusion checker (hard fork of Apiiro's Dependency Combobulator)",
        epilog='Hard fork of Apiiro/combobulator by cognitivegears',
        add_help=True)
    
    parser.add_argument("-t", "--type",
                        dest="package_type",
                        help="Package Manager Type, i.e: npm, PyPI, maven",
                        action="store", type=str, 
                        choices=Constants.SUPPORTED_PACKAGES,
                        required=True)

    input_group = parser.add_mutually_exclusive_group(required=True)
    input_group.add_argument("-l", "--load_list",
                        dest="LIST_FROM_FILE",
                        help="Load list of dependencies from a file",
                        action="append", type=str,
                        default=[])
    input_group.add_argument("-d", "--directory",
                    dest="FROM_SRC",
                    help="Extract dependencies from local source repository",
                    action="append",
                    type=str)
    input_group.add_argument("-p", "--package",
                            dest="SINGLE",
                            help="Name a single package.",
                            action="append", type=str)

    output_group = parser.add_mutually_exclusive_group(required=False)
    output_group.add_argument("-c", "--csv",
        dest="CSV",
        help="Export packages properties onto CSV file",
                    action="store", type=str)
    output_group.add_argument("-j", "--json",
        dest="JSON",
        help="Export packages properties onto JSON file",
        action="store", type=str)

    parser.add_argument("-a", "--analysis",
        dest="LEVEL",
        help="Required analysis level - compare (comp), heuristics (heur) (default: compare)",
                    action="store", default="compare", type=str,
                    choices=Constants.LEVELS)
    parser.add_argument("--loglevel",
                        dest="LOG_LEVEL",
                        help="Set the logging level",
                        action="store",
                        type=str,
                        choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'],
                        default='INFO')
    parser.add_argument("--logfile",
                        dest="LOG_FILE",
                        help="Log output file",
                        action="store",
                        type=str)
    parser.add_argument("-r", "--recursive",
                        dest="RECURSIVE",
                        help="Recursively scan directories when scanning from source.",
                        action="store_true")
    parser.add_argument("--error-on-warnings",
                        dest="ERROR_ON_WARNINGS",
                        help="Exit with a non-zero status code if warnings are present.",
                        action="store_true")
    parser.add_argument("-q", "--quiet",
                        dest="QUIET",
                        help="Do not output to console.",
                        action="store_true")

    return parser.parse_args()
